:- dynamic testsetBase/1.
:- consult(simplify).
:- op(900,fy,not).

%---------------------------------------------------------------------------
% Path of the MONA executable on your machine

monaExec('/usr/local/mona-1.4/mona').

%---------------------------------------------------------------------------
% CPU time limit for the execution of MONA (in seconds).
% The limit will be imposed by `ulimit'. 

timeLimit(1000).

%---------------------------------------------------------------------------
% Various options
% The performance of MONA does not seem to be effected by
% any kind of pre-simplification of PLTL formulae we have
% tried.

simplifyOpt(no).
simplifications([]).
orderingOpt(normal).
sortOpt([]).

%---------------------------------------------------------------------------
% testForm(+FORMULA)
% test the satisfiability of the formula FORMULA 

testForm(Term) :-
	pltlSimplification(Term,Term2),
	MONA_Input_File = '/var/tmp/mona.in',
	MONA_Output_File = '/var/tmp/mona.out',
	tell(MONA_Input_File),
	printFormulaMONA(Term2),
	told,
	format("Starting computation on ~a~n",[MONA_Input_File]),
	format("using - MONA 1.4~n",[]),
	format("writing protocol to ~a.~n",[MONA_Output_File]),
	ttyflush,
	!,
	monaExec(MONA),
        timeLimit(TL),
	graciousCall(exec,['ulimit -t ',TL,'; /usr/bin/time -f "%Uuser %Ssystem %ewall %Mmem" ',MONA,' -q -t ',MONA_Input_File,' > ',MONA_Output_File,' 2>&1']),
	systemCall(system,['cat ',MONA_Output_File]),
	ttyflush.
	
%---------------------------------------------------------------------------
% testFile(+FILENAME)
% test the satisfiability of the formula contained in the file
% FILENAME. 

testFile(File) :-
	see(File),
	read(Term),
	seen,
	testForm(Term),
	!.

%---------------------------------------------------------------------------
% printFormulaMONA(+TERM)
% print term TERM using the MONA syntax for WS1S formulae.
%

printFormulaMONA(F) :-
	write('var1 b,e;'),nl,
	getSignature(F,Sig),
	printVarDeclMONA(Sig), nl, nl,
	write('#next predicate'), nl,
	write('pred next(var1 x, var1 y) = '), nl,
        write('     (((x < e) & (y = x + 1)) | ((x = e) & (y = b + 1)));'),nl,
	write('#connect predicate'), nl,
	write('pred conn(var1 x, var1 y) = '), nl,
	write('(ex2 S: (x in S & y in S & '), nl,
	write('         (all1 y1: (all1 z1: ((y1 in S & next(y1,z1)) => z1 in S))) &'),nl,
 	write('         (all2 T: ((x in T & '),nl,
        write('                   (all1 y2: (all1 z2: ((y2 in T & next(y2,z2)) => z2 in T)))) '),
        write(' => (S sub T)))));'), nl,
	printFormMONA(F,0), write(';'), nl,
	!.

printVarDeclMONA([]) :-
	!.
printVarDeclMONA([V|VL]) :-
	write('var2 '),
	printVarDeclMONA1([V|VL]),
	write(';').

printVarDeclMONA1([]) :-
	!.
printVarDeclMONA1([V]) :-
	write(V).
printVarDeclMONA1([V1,V2|VL]) :-
	write(V1),
	write(', '),
	printVarDeclMONA1([V2|VL]).

printFormMONA(always(C),X) :-
	gensym(x,Y),
	format("(all1 ~a: (conn(~a,~a) => ",[Y,X,Y]),
	printFormMONA(C,Y),
	format("))~n",[]).
printFormMONA(sometime(C),X) :-
	gensym(x,Y),
	format("(ex1 ~a: (conn(~a,~a) & ",[Y,X,Y]),
	printFormMONA(C,Y),
	format("))",[]).
printFormMONA(next(C),X) :-
	gensym(x,Y),
	format("(ex1 ~a: (next(~a,~a) & ",[Y,X,Y]),
	printFormMONA(C,Y),
	format("))",[]).
printFormMONA(and(L),X) :-
	write('('),
	printFormsMONA(' & ',L,X),
	write(')').
printFormMONA(or(L),X) :-
	write('('),
	printFormsMONA(' | ',L,X),
	write(')').
printFormMONA(not(A),X) :-
	write('~( '),
	printFormMONA(A,X),
	write(')').
printFormMONA(A,X) :-
	format("(~a in ~a)",[X,A]).
printFormsMONA(_Op,[T1],X) :-
	printFormMONA(T1,X).
printFormsMONA(Op,[T1,T2|TL],X) :-
	printFormMONA(T1,X),
	write(' '),
	write(Op),
	write(' '),
	printFormsMONA(Op,[T2|TL],X).

generateAtom([],AtomString,Atom) :-
	atom_chars(Atom,AtomString).
generateAtom([A|L],AtomString1,Atom) :-
	(number(A) ->
	    number_chars(A,AC)
	;
	    (atomic(A) ->
		atom_chars(A,AC)
	    ;
		AC = A
	    )
	),
	append(AtomString1,AC,AtomString2),
	generateAtom(L,AtomString2,Atom).

graciousCall(CallPredicate,Command) :-
	once(systemCall(Command,[],CallPredicate)),
	!.
graciousCall(_,_) :-
	!.

systemCall(CallPredicate,Command) :-
	once(systemCall(Command,[],CallPredicate)),
	!.
systemCall(_CallPredicate,Command) :-
	write('SystemCall of '),
	write(Command),
	write(' failed.'),
	nl,
	!.

systemCall([],CommandString,_) :-
	atom_chars(Command,CommandString),
	!,
	shell(Command).
systemCall([],CommandString,shell(Result)) :-
	atom_chars(Command,CommandString),
	!,
	shell(Command,Result).
systemCall([A|L],CommandString1,CallPredicate) :-
	(number(A) ->
	    number_chars(A,AC)
	;
	    (atomic(A) ->
		atom_chars(A,AC)
	    ;
		AC = A
	    )
	),
	append(CommandString1,AC,CommandString2),
	systemCall(L,CommandString2,CallPredicate).



pltlSimplification(Term,Term3) :-
	(simplifyOpt(yes) ->
	    simplify_pltl(Term,Term1)
	;
	    Term1 = Term
	),
	(orderingOpt(reverse) ->
	    (Term1 = and(L) ->
		reverse(L,L2),
		Term3 = and(L2)
	    ;
		(Term1 = or([and([C1,and(L1)]),and([C2,and(L2)])]) ->
		    reverse([C1|L1],RL1),
		    reverse([C2|L2],RL2),
		    Term3 = or([and(RL1),and(RL2)])
		;
		    (Term1 = or([and(L1),and(L2)]) ->
			reverse(L1,RL1),
			reverse(L2,RL2),
			Term3 = or([and(RL1),and(RL2)])
		    ;
			Term3 = Term1
		    )
		)
	    )
	;
	    Term3 = Term1
	),
	!.

getSignature([],[]) :-
        !.
getSignature(Term,[Term]) :-
        atomic(Term),
        !.
getSignature(Term,Sig) :-
        Term =.. [_|Terms],
        getSignatures(Terms,Sig).
getSignatures([],[]).
getSignatures([T1|TL1],Sig) :-
        getSignature(T1,Sig1),
        getSignatures(TL1,Sig2),
        oset_union(Sig1,Sig2,Sig).

start_message :-
        format("PLTL-to-MONA Translator Version 1.0~n~n",[]),
        write('Copyright (c) 2001 Ullrich Hustadt, University of Liverpool.'), nl,
        write('Copy policy: GPL-2 (see www.gnu.org)'), nl,
        write('Implemented in SWI-Prolog (Version 4.0.8)'), nl,
        write('Copyright (c) 1990-2000 University of Amsterdam.'), nl,
        write('Copy policy: GPL-2 (see www.gnu.org)'), nl, nl,
        write('For help on the translor, use ?- help(pltl).'), nl,
        write('For help on SWI-Prolog, use ?- help(Topic). or ?- apropos(Word).'), nl, nl.

help(pltl) :-
        format("PLTL-to-MONA Translator Version 1.0~n~n",[]),
        format("INPUT SYNTAX FOR PLTL FORMULAE:~n",[]),
        format("FORMULA             ::= ATOM | and(LIST_OF_FORMULAE) | or(LIST_OF_FORMULAE) | not(FORMULA) | next(FORMULA) | always(FORMULA) | sometime(FORMULA)~n",[]),
        format("LIST_OF_FORMULAE    ::= [FORMULA,...,FORMULA]~n",[]),
	format("ATOM                ::= [a-z][a-z0-9]+~n",[]),
        format("EXAMPLE OF A FORMULA:~n",[]),
        format("and([or([p0,not(p1)]),~n",[]),
        format("     always(or([p1,not(p0)])),~n",[]),
        format("     always(or([not(p0), next(not p4), next(p1)])),~n",[]),
        format("     always(or([not(p1), sometime(p2)]))]),~n",[]),
        format("USAGE:~n",[]),
        format("- testFile(FILENAME).~n",[]),
        format("  FILENAME is the name of a file containg a FORMULA followed by `.'~n",[]),
        format("  FORMULA will be tested for satisfiability by using a translation~n",[]),
	format("  MONA input syntax and execution of MONA.~n",[]),
        format("- testForm(FORMULA).~n",[]),
        format("  FORMULA will be tested for satisfiability by using a translation~n",[]),
	format("  MONA input syntax and execution of MONA.",[]),
	!.

:- at_initialization(start_message).
:- qsave_program('PLTLmona').
